(* Algorithme d'analyse de flot *)

let succs x s = set_map snd (filter_set (fun (y,_) -> y = x) s);;
let preds x s = set_map fst (filter_set (fun (_,y) -> y = x) s);;

let t12_sub x expr e =
 let e2 = if is_k x then e else filter_set (fun (y,_) -> y <> x) e in
 match expr with
    Base(Var(y)) -> union e2 (product (singleton x) (succs y e))
  | Adress(y) -> union e2 (singleton (x,y))
  | Content(y) -> let pointes = succs y e in
                  if pointes = empty_set then failwith ""
                  else union e2 (product (singleton x)
                        (unionset (set_map (fun z -> succs z e) pointes)))
  | New(Var(y),ne) -> union e2 (union (singleton (x,ne))
                       (product (singleton ne) (succs y e)))
  | New(Cte(_),ne) -> union e2 (singleton (x,ne))
  | _ -> e2;;

let t12 e = function
   Command(Variable(x),expr) -> t12_sub x expr e
 | Command(Deref(x),expr) -> let succs = succs x e in
       if succs = empty_set then failwith ""
       else unionset (set_map (fun y -> t12_sub y expr e) succs)
 | Test(_) -> e
 | Skip -> e
 | NonLie -> empty_set;;

let t1 es = function
   Command(Variable(x),expr) -> set_map (t12_sub x expr) es
 | Command(Deref(x),expr) -> 
         unionset (set_map (fun e -> let succs = succs x e in
         if succs = empty_set then failwith ""
         else set_map (fun y -> t12_sub y expr e) succs) es)
 | Test(_) -> es
 | Skip -> es
 | NonLie -> empty_set;;

let analyse bottom init us t p =
 let g = graph_of_prog p in
 let ng = Array.length g in
 let tab = Array.make ng bottom in
 tab.(0) <- init;
 let convergence = ref false in
 while not (!convergence) do
  convergence := true;
  for i = 0 to ng - 1 do
   let old = tab.(i)
   and news = List.map (fun j -> t tab.(j) g.(j).(i)) (list ng) in
   tab.(i) <- us (if i = 0 then init :: news else news);
   if old <> tab.(i) then convergence := false
  done;
 done;
 tab;;

(* pour les programmes sans "new" *)

let analyse12 = analyse empty_set empty_set unions t12;;
let analyse1 = analyse empty_set (singleton empty_set) unions t1;;
let analyse1then2 p = Array.map unionset (analyse1 p);;

(* pour les programmes avec "new" *)

let nanalyse12 p = analyse12 (deplie p);;
let nanalyse1 p = analyse1 (deplie p);;
let nanalyse1then2 p = analyse1then2 (deplie p);;
