(* Construction des langages rationnels associs  un graphe *)

type lrat =
    Union of lrat list
  | Concat of lrat list
  | Star of lrat
  | Vr of ident;;

let rec simplify = function
   (Union [x]) -> simplify x
 | (Concat [x]) -> simplify x
 | (Star (Concat [])) -> Concat []
 | (Star (Union [])) -> Union []
 | (Star (Union l)) -> let l2 = List.map simplify l in
                       let l3 = filter (fun x -> x <> Concat []) l2 in
                       if l3 <> l then simplify (Star (Union l3))
                                  else Star (Union l)
 | (Union l) -> Union (List.map simplify l)
 | (Concat l) -> let l2 = List.map simplify l in
                 let l3 = filter (fun x -> x <> Concat []) l2 in
                 if l3 <> l then simplify (Concat l3) else Concat l
 | (Star l) -> Star (simplify l)
 | (Vr x) -> Vr x;;

let unionrat l1 l2 = simplify (match (l1,l2) with
   (Union u1,Union u2) -> Union (merge u1 u2)
 | (Union u1,_) -> Union (merge u1 [l2])
 | (_,Union u2) -> Union (merge [l1] u2)
 | _ -> if l1 = l2 then l1 else Union [l1;l2]);;

let concatrat l1 l2 = simplify (match (l1,l2) with
   (Union [],_) -> Union []
 | (_,Union []) -> Union []
 | (Concat c1,Concat c2) -> Concat (c1 @ c2)
 | (Concat c1,_) -> Concat (c1 @ [l2])
 | (_,Concat c2) -> Concat (l1 :: c2)
 | _ -> Concat [l1;l2]);;

let rec inverserat = function
   (Union l) -> Union (List.map inverserat l)
 | (Concat l) -> Concat (List.map inverserat (List.rev l))
 | (Star l) -> Star (inverserat l)
 | (Vr x) -> Vr x;;

let construit_tab g vars =
 let var k = List.nth vars k in
 let n = List.length vars in
 let r = Array.make_matrix (n+1) (n+1) (Union []) in
 for j = 1 to n do r.(0).(j) <- Vr (var (j-1)) done;
 for i = 1 to n do
  for j = 1 to n do
   if in_set (var (i-1),var (j-1)) g then r.(i).(j) <- Vr "*"
  done
 done;
 for i = 0 to n do r.(i).(i) <- unionrat r.(i).(i) (Concat []) done;
 r;;

let iter n k r =
 let r2 = Array.make_matrix n n (Union []) in
 for i = 0 to n-1 do
  for j = 0 to n-1 do
   r2.(i).(j) <- unionrat r.(i).(j) (concatrat r.(i).(k)
                                    (concatrat (Star r.(k).(k)) r.(k).(j)))
  done
 done;
 r2;;

let classes p g =
 let vars = set_to_list (varset p) in
 let r = ref (construit_tab g vars) in
 let n = Array.length !r in
 for k = 1 to n-1 do r := iter n k !r done;
 Array.map inverserat (Array.sub !r.(0) 1 (n-1));;

(* Attention: l'etoile est note "+" pour ne pas la confondre avec
   le "*" du drfrencement *)

let rec print_lrat = function
   (Union []) -> ()
 | (Union l) -> (print_char '('; print_lrat (List.hd l); List.iter
                 (fun x -> print_string " U "; print_lrat x) (List.tl l);
                 print_char ')')
 | (Concat []) -> print_string "eps"
 | (Concat l) -> List.iter print_lrat l
 | (Star (Union l)) -> (print_lrat (Union l); print_char '+')
 | (Star l) -> (print_char '('; print_lrat l; print_string ")+")
 | (Vr x) -> print_string x;;

let print_classes p g =
 Array.iter (fun x -> print_lrat x; print_newline ()) (classes p g);;
