(* Cration du graphe de flot de controle *)

type arc =
    Command of location * expr
  | Test of expr
  | Skip
  | NonLie;;

let rec nb_sommets p =
 let rec nb_aux = function
    Affect(_,_) -> 1
  | SSkip -> 1
  | If(_,i1,i2) -> 3 + nb_aux i1 + nb_aux i2
  | While(_,i) -> 2 + nb_aux i
  | Sequence(i1,i2) -> nb_aux i1 + nb_aux i2
 in 1 + nb_aux p;;

let inverse_test = function
    Equal(e1,e2) -> Diff(e1,e2)
  | Diff(e1,e2) -> Equal(e1,e2)
  | Less(e1,e2) -> Greater(e1,e2)  (* attention: non < == >= *)
  | Greater(e1,e2) -> Less(e1,e2)
  | _ -> failwith "";;

let graph_of_prog p =
 let n = nb_sommets p in
 let g = Array.make_matrix n n NonLie in
 let rec parcours s = function
    Affect(var,e) -> begin g.(s).(s+1) <- Command(var,e);
                           s + 1
                     end
  | SSkip -> begin g.(s).(s+1) <- Skip;
                   s + 1
             end
  | If(e,i1,i2) -> begin g.(s).(s+1) <- Test(e);
                         let t = parcours (s+1) i1 in
                         g.(s).(t+1) <- Test(inverse_test e);
                         let u = parcours (t+1) i2 in
                         g.(t).(u+1) <- Skip;
                         g.(u).(u+1) <- Skip;
                         u + 1
                   end
  | While(e,i) -> begin g.(s).(s+1) <- Test(e);
                        let t = parcours (s+1) i in
                        g.(t).(s) <- Skip;
                        g.(s).(t+1) <- Test(inverse_test e);
                        t + 1
                  end
  | Sequence(i1,i2) -> parcours (parcours s i1) i2
 in parcours 0 p;
    g;;
