(* Parseur *)

open Genlex;;

let lexer = make_lexer [";"; "skip"; "if"; "then"; "else"; ":="; "while"; "do";
                        "("; ")"; "+"; "-"; "*"; "/"; "="; "<>"; "<"; ">"; 
                        "or"; "and"; "not"; "&"; "new"];;

let read_baseexpr = parser
  [< 'Ident var >] -> Var(var)
| [< 'Int n >] -> Cte(n)
| [< 'Kwd "-";'Int n >] -> Cte(-n);;

let read_expr2 expr1 = parser
  [< 'Kwd "+"; expr2 = read_baseexpr >] -> Plus(expr1,expr2)
| [< 'Kwd "-"; expr2 = read_baseexpr >] -> Minus(expr1,expr2)
| [< 'Kwd "*"; expr2 = read_baseexpr >] -> Mul(expr1,expr2)
| [< 'Kwd "/"; expr2 = read_baseexpr >] -> Div(expr1,expr2)
| [< 'Kwd "="; expr2 = read_baseexpr >] -> Equal(expr1,expr2)
| [< 'Kwd "<>"; expr2 = read_baseexpr >] -> Diff(expr1,expr2)
| [< 'Kwd "<"; expr2 = read_baseexpr >] -> Less(expr1,expr2)
| [< 'Kwd ">"; expr2 = read_baseexpr >] -> Greater(expr1,expr2)
| [< >] -> Base(expr1);;

let read_expr = parser
  [< 'Kwd "&"; 'Ident var >] -> Adress(var)
| [< 'Kwd "*"; 'Ident var >] -> Content(var)
| [< 'Kwd "new"; expr = read_baseexpr >] -> New(expr,var (new_num ()))
| [< expr1 = read_baseexpr; expr = read_expr2 expr1 >] -> expr;;

let read_location = parser
  [< 'Kwd "*"; 'Ident var >] -> Deref(var)
| [< 'Ident var >] -> Variable(var);;

let rec read_prog = parser
  [< 'Kwd "("; prog = read_sequence; 'Kwd ")" >] -> prog

and read_sequence st =
 let instr = read_instr st in
 match st with parser
  [< 'Kwd ";"; suite = read_sequence >] -> Sequence(instr,suite)
| [< >]                                 -> instr

and read_instr = parser
  [< 'Kwd "if"; b = read_expr; 'Kwd "then"; prog1 = read_prog;
     'Kwd "else"; prog2 = read_prog >] -> If(b,prog1,prog2)
| [< 'Kwd "while"; b = read_expr; 'Kwd "do"; prog = read_prog >]
    -> While(b,prog)
| [< 'Kwd "skip" >] -> SSkip
| [< var = read_location; 'Kwd ":="; expr2 = read_expr >]
    -> Affect(var,expr2);;

let read() = read_prog (lexer (Stream.of_channel stdin));;
