(* Module d'implmentation d'un type "ensemble"
   Gagnerait beaucoup en rapidit  tre implment sous forme
   d'arbres binaires *)

type 'a set = Set of 'a list;;

let empty_set = Set [];;
let singleton x = Set [x];;

let add_to_set x (Set l) =
 let rec loop = function
    (y::ys) -> if x = y then y::ys
                      else if x < y then x::y::ys
                      else y::loop ys
  | [] -> [x]
 in Set (loop l);;

let list_to_set l = List.fold_right add_to_set l empty_set;;
let set_to_list (Set l) = l;;

let set_fold f e (Set l) =
 let rec loop e = function
    (x::xs) -> loop (f e x) xs
  | [] -> e
 in loop e l;;

let union (Set l1) (Set l2) =
 let rec loop = function
    (x::xs,y::ys) -> if x = y then x::loop (xs,ys)
                     else if x < y then x::loop (xs,y::ys)
                     else y::loop (x::xs,ys)
  | ([],l) -> l
  | (l,[]) -> l
 in Set (loop (l1,l2));;

let unions l = List.fold_left union empty_set l;;
let unionset (Set l) = unions l;;

let in_set x (Set l) =
 let rec loop = function
    (y::ys) -> x = y || (x > y && loop ys)
  | [] -> false
 in loop l;;

let subset (Set l1) (Set l2) =
 let rec loop = function
    (x::xs,y::ys) -> if x = y then loop (xs,ys)
                     else if x < y then false
                     else loop (x::xs,ys)
  | ([],l) -> true
  | (l,[]) -> false
 in loop (l1,l2);; 

let intersection (Set l1) (Set l2) =
 let rec loop = function
    (x::xs,y::ys) -> if x = y then x::loop (xs,ys)
                     else if x < y then loop (xs,y::ys)
                     else loop (x::xs,ys)
  | _ -> []
 in Set (loop (l1,l2));;

let intersections = function
   [] -> failwith "Liste vide"
 | (s::ss) -> List.fold_left intersection s ss;;
let intersectionset (Set l) = intersections l;;

let minus (Set l1) (Set l2) =
 let rec loop = function
    (x::xs,y::ys) -> if x = y then loop (xs,ys)
                     else if x < y then x::loop (xs,y::ys)
                     else loop (x::xs,ys)
  | (l,[]) -> l
  | ([],l) -> []
 in Set (loop (l1,l2));;

let product (Set l1) (Set l2) =
 let prod_un x = List.map (fun y -> (x,y)) l2
 in Set (List.concat (List.map prod_un l1));;

let filter_set p (Set l) =
 let rec loop = function
    (x::xs) -> if p x then x::loop xs else loop xs
  | [] -> []
 in Set (loop l);;

let set_map f (Set l) = list_to_set (List.map f l);;
let set_iter f (Set l) = List.iter f l;;

let set_of_subsets st =
 set_fold (fun s x -> union s (set_map (add_to_set x) s))
          (singleton empty_set) st;;

let print_set f = function
   (Set []) -> print_string "{}"
 | (Set (x::xs)) -> (print_char '{'; f x;
                     List.iter (fun y -> print_string "; "; f y) xs;
                     print_char '}');;
